################################################################################
# Copyright (c) 2014-2022, 2025, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for RFI flagging wrappers.

The backgrounding and thresholding have separate tests - this module just tests
that they can be glued together properly.
"""

import numpy as np
import pytest

from katsdpsigproc.abc import AbstractCommandQueue, AbstractContext
from katsdpsigproc.rfi import device, host

from .. import complex_normal

_vis: np.ndarray
_spikes: np.ndarray
_input_flags: np.ndarray


@pytest.fixture(autouse=True)
def setup() -> None:
    global _vis, _spikes, _input_flags
    shape = (117, 131)
    # Use a fixed seed to make the test repeatable
    rs = np.random.RandomState(seed=1)
    _vis = complex_normal(rs, size=shape)
    # Pick 1/16 of samples to be RFI
    # Give them random amplitude from a range, and random phase
    _spikes = rs.random_sample(shape) < 1.0 / 16.0
    _spikes = _spikes.astype(np.uint8)
    rfi_amp = rs.random_sample(shape) * 20.0 + 50.0
    rfi_phase = rs.random_sample(shape) * (2j * np.pi)
    rfi = rfi_amp * np.exp(rfi_phase)
    _vis += _spikes * rfi
    _vis = _vis.astype(np.complex64)
    _input_flags = (rs.random_sample(shape) < 1.0 / 16.0).astype(np.uint8) * 2


def test_flagger_host() -> None:
    global _vis, _spikes, _input_flags
    background = host.BackgroundMedianFilterHost(13)
    noise_est = host.NoiseEstMADHost()
    threshold = host.ThresholdSimpleHost(11.0)
    flagger = host.FlaggerHost(background, noise_est, threshold)
    flags = flagger(_vis)
    np.testing.assert_equal(_spikes, flags)
    # Test again, this time with channel flags
    flags = flagger(_vis, _input_flags[:, 0])
    input_flags = np.broadcast_to(_input_flags[:, 0:1], _vis.shape)
    expected = np.where(input_flags, 0, _spikes)
    np.testing.assert_equal(expected, flags)
    # And again with full input flags
    flags = flagger(_vis, _input_flags)
    expected = np.where(_input_flags, 0, _spikes)
    np.testing.assert_equal(expected, flags)


def check_flagger_device(
    use_flags: device.BackgroundFlags,
    transpose_noise_est: bool,
    transpose_threshold: bool,
    context: AbstractContext,
    command_queue: AbstractCommandQueue,
) -> None:
    global _vis, _spikes, _input_flags
    background = device.BackgroundMedianFilterDeviceTemplate(context, 13, use_flags=use_flags)
    if transpose_noise_est:
        noise_est: device.AbstractNoiseEstDeviceTemplate = device.NoiseEstMADTDeviceTemplate(
            context, 1024
        )
    else:
        noise_est = device.NoiseEstMADDeviceTemplate(context, tuning={"wgsx": 8, "wgsy": 8})
    threshold = device.ThresholdSimpleDeviceTemplate(
        context, transpose_threshold, tuning={"wgsx": 8, "wgsy": 8}
    )
    flagger_device = device.FlaggerDeviceTemplate(background, noise_est, threshold)
    flagger = device.FlaggerHostFromDevice(
        flagger_device, command_queue, threshold_args=dict(n_sigma=11.0)
    )
    if use_flags == device.BackgroundFlags.CHANNEL:
        flags = flagger(_vis, _input_flags[:, 0])
        input_flags = np.broadcast_to(_input_flags[:, 0:1], _vis.shape)
        expected = np.where(input_flags, 0, _spikes)
        np.testing.assert_equal(expected, flags)
    elif use_flags == device.BackgroundFlags.FULL:
        flags = flagger(_vis, _input_flags)
        expected = np.where(_input_flags, 0, _spikes)
        np.testing.assert_equal(expected, flags)
    else:
        flags = flagger(_vis)
        np.testing.assert_equal(_spikes, flags)


def test_flagger_device(context: AbstractContext, command_queue: AbstractCommandQueue) -> None:
    check_flagger_device(device.BackgroundFlags.NONE, False, False, context, command_queue)


def test_flagger_device_transpose_noise_est(
    context: AbstractContext, command_queue: AbstractCommandQueue
) -> None:
    """Test device flagger with a transposed noise estimator."""
    check_flagger_device(device.BackgroundFlags.CHANNEL, True, False, context, command_queue)


def test_flagger_device_transpose_threshold(
    context: AbstractContext, command_queue: AbstractCommandQueue
) -> None:
    """Test device flagger with a transposed thresholder."""
    check_flagger_device(device.BackgroundFlags.FULL, False, True, context, command_queue)


def test_flagger_device_transpose_both(
    context: AbstractContext, command_queue: AbstractCommandQueue
) -> None:
    """Test device flagger with a transposed noise estimator and thresholder."""
    check_flagger_device(device.BackgroundFlags.NONE, True, True, context, command_queue)



################################################################################
# Copyright (c) 2014-2021, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Transpose 2D arrays on a device."""

from typing import Any, Callable, Mapping, Optional, Tuple, cast

import numpy as np
from typing_extensions import TypedDict

try:
    from numpy.typing import DTypeLike
except ImportError:
    DTypeLike = Any  # type: ignore

from . import accel, tune
from .abc import AbstractCommandQueue, AbstractContext


class _TuningDict(TypedDict):
    block: int
    vtx: int
    vty: int


class TransposeTemplate:
    """Kernel for transposing a 2D array of data.

    Parameters
    ----------
    context
        Context for which kernels will be compiled
    dtype
        Type of data elements
    ctype
        Type (in C/CUDA, not numpy) of data elements
    tuning
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - block: number of workitems per workgroup in each dimension
        - vtx, vty: elements per workitem in each dimension
    """

    autotune_version = 1

    def __init__(
        self,
        context: AbstractContext,
        dtype: DTypeLike,
        ctype: str,
        tuning: Optional[_TuningDict] = None,
    ) -> None:
        self.context = context
        self.dtype: np.dtype = np.dtype(dtype)
        self.ctype = ctype
        if tuning is None:
            tuning = self.autotune(context, dtype, ctype)
        self._block = tuning["block"]
        self._tilex = tuning["block"] * tuning["vtx"]
        self._tiley = tuning["block"] * tuning["vty"]
        self.program = accel.build(
            context,
            "transpose.mako",
            {
                "block": tuning["block"],
                "vtx": tuning["vtx"],
                "vty": tuning["vty"],
                "ctype": ctype,
            },
        )

    @classmethod
    @tune.autotuner(test={"block": 8, "vtx": 2, "vty": 3})
    def autotune(cls, context: AbstractContext, dtype: DTypeLike, ctype: str) -> _TuningDict:
        queue = context.create_tuning_command_queue()
        in_shape = (2048, 2048)
        out_shape = (2048, 2048)
        in_data = accel.DeviceArray(context, in_shape, dtype=dtype)
        out_data = accel.DeviceArray(context, out_shape, dtype=dtype)

        def generate(block: int, vtx: int, vty: int) -> Optional[Callable[[int], float]]:
            local_mem = (block * vtx + 1) * (block * vty) * np.dtype(dtype).itemsize
            if local_mem > 32768:
                # Skip configurations using lots of lmem
                raise RuntimeError("too much local memory")
            fn = cls(context, dtype, ctype, {"block": block, "vtx": vtx, "vty": vty}).instantiate(
                queue, in_shape
            )
            fn.bind(src=in_data, dest=out_data)
            return tune.make_measure(queue, fn)

        return cast(
            _TuningDict,
            tune.autotune(generate, block=[4, 8, 16, 32], vtx=[1, 2, 3, 4], vty=[1, 2, 3, 4]),
        )

    def instantiate(
        self,
        command_queue: AbstractCommandQueue,
        shape: Tuple[int, int],
        allocator: Optional[accel.AbstractAllocator] = None,
    ) -> "Transpose":
        return Transpose(self, command_queue, shape, allocator)


class Transpose(accel.Operation):
    """Concrete instance of :class:`TransposeTemplate`.

    .. rubric:: Slots

    **src**
        Input

    **dest**
        Output
    """

    def __init__(
        self,
        template: TransposeTemplate,
        command_queue: AbstractCommandQueue,
        shape: Tuple[int, int],
        allocator: Optional[accel.AbstractAllocator] = None,
    ):
        super().__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel("transpose")
        self.shape = shape
        self.slots["src"] = accel.IOSlot(shape, template.dtype)
        self.slots["dest"] = accel.IOSlot((shape[1], shape[0]), template.dtype)

    def _run(self) -> None:
        src = self.buffer("src")
        dest = self.buffer("dest")
        # Round up to number of blocks in each dimension
        in_row_tiles = accel.divup(src.shape[0], self.template._tiley)
        in_col_tiles = accel.divup(src.shape[1], self.template._tilex)
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                dest.buffer,
                src.buffer,
                np.int32(src.shape[0]),
                np.int32(src.shape[1]),
                np.int32(dest.padded_shape[1]),
                np.int32(src.padded_shape[1]),
            ],
            global_size=(
                in_col_tiles * self.template._block,
                in_row_tiles * self.template._block,
            ),
            local_size=(self.template._block, self.template._block),
        )

    def parameters(self) -> Mapping[str, Any]:
        return {
            "dtype": self.template.dtype,
            "ctype": self.template.ctype,
            "shape": self.slots["src"].shape,  # type: ignore
        }



import numpy as np

import katsdpsigproc.accel
import katsdpsigproc.fill
import katsdpsigproc.reduce


class FillReduceTemplate:
    def __init__(self, context):
        self.fill = katsdpsigproc.fill.FillTemplate(context, np.float32, "float")
        self.hreduce = katsdpsigproc.reduce.HReduceTemplate(
            context, np.float32, "float", "a+b", "0.0f"
        )

    def instantiate(self, queue, shape):
        return FillReduce(self, queue, shape)


class FillReduce(katsdpsigproc.accel.OperationSequence):
    def __init__(self, template, queue, shape):
        self.fill = template.fill.instantiate(queue, shape)
        self.hreduce = template.hreduce.instantiate(queue, shape)
        operations = [("fill", self.fill), ("hreduce", self.hreduce)]
        compounds = {"src": ["fill:data", "hreduce:src"], "dest": ["hreduce:dest"]}
        super().__init__(queue, operations, compounds)
        self.template = template

    def __call__(self, fill_value):
        self.fill.set_value(fill_value)
        super().__call__()


ctx = katsdpsigproc.accel.create_some_context()
queue = ctx.create_command_queue()
op_template = FillReduceTemplate(ctx)
op = op_template.instantiate(queue, (10, 5))
op(42)
print(op.buffer("dest").get(queue))



################################################################################
# Copyright (c) 2014-2021, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Perform on-device percentile calculation of 2D arrays."""
# see scripts/percentiletest.py for an example

from typing import Any, Callable, Mapping, Optional, Tuple, cast

import numpy as np
from typing_extensions import TypedDict

from . import accel, tune
from .abc import AbstractCommandQueue, AbstractContext


class _TuningDict(TypedDict):
    size: int
    wgsy: int


class Percentile5Template:
    """Kernel for calculating percentiles of a 2D array of data.

    5 percentiles [0,100,25,75,50] are calculated per row (along columns, independently per row).
    The lower percentile element, rather than a linear interpolation is chosen.
    WARNING: assumes all values are positive.

    Parameters
    ----------
    context
        Context for which kernels will be compiled
    max_columns
        Maximum number of columns
    is_amplitude
        If true, the inputs are scalar amplitudes; if false, they are complex
        numbers and the answers are computed on the absolute values
    tuning
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - size: number of workitems per workgroup along each row
        - wgsy: number of workitems per workgroup along each column
    """

    autotune_version = 8

    def __init__(
        self,
        context: AbstractContext,
        max_columns: int,
        is_amplitude: bool = True,
        tuning: Optional[_TuningDict] = None,
    ) -> None:
        self.context = context
        self.max_columns = max_columns
        self.is_amplitude = is_amplitude

        if tuning is None:
            tuning = self.autotune(context, max_columns, is_amplitude)
        self.size = tuning["size"]
        self.wgsy = tuning["wgsy"]
        self.vt = accel.divup(max_columns, tuning["size"])
        self.program = accel.build(
            context,
            "percentile.mako",
            {
                "size": self.size,
                "wgsy": self.wgsy,
                "vt": self.vt,
                "is_amplitude": self.is_amplitude,
            },
        )

    @classmethod
    @tune.autotuner(test={"size": 64, "wgsy": 4})
    def autotune(
        cls, context: AbstractContext, max_columns: int, is_amplitude: bool
    ) -> _TuningDict:
        queue = context.create_tuning_command_queue()
        in_shape = (4096, max_columns)
        rs = np.random.RandomState(seed=1)
        if is_amplitude:
            host_data: np.ndarray = rs.uniform(size=in_shape).astype(np.float32)
        else:
            host_data = rs.standard_normal(in_shape) + 1j * rs.standard_normal(in_shape)
            host_data = host_data.astype(np.complex64)

        def generate(size: int, wgsy: int) -> Callable[[int], float]:
            if size * wgsy < 32 or size * wgsy > 1024:
                raise RuntimeError("work group size is unnecessarily large or small, skipping")
            if max_columns > size * 256:
                raise RuntimeError("too many columns")
            fn = cls(context, max_columns, is_amplitude, {"size": size, "wgsy": wgsy}).instantiate(
                queue, in_shape
            )
            inp = fn.slots["src"].allocate(fn.allocator)
            fn.slots["dest"].allocate(fn.allocator)
            inp.set(queue, host_data)
            return tune.make_measure(queue, fn)

        return cast(
            _TuningDict,
            tune.autotune(
                generate,
                size=[8, 16, 32, 64, 128, 256, 512, 1024],
                wgsy=[1, 2, 4, 8, 16, 32],
            ),
        )

    def instantiate(
        self,
        command_queue: AbstractCommandQueue,
        shape: Tuple[int, int],
        column_range: Optional[Tuple[int, int]] = None,
        allocator: Optional[accel.AbstractAllocator] = None,
    ) -> "Percentile5":
        return Percentile5(self, command_queue, shape, column_range, allocator)


class Percentile5(accel.Operation):
    """Concrete instance of :class:`PercentileTemplate`.

    .. warning::

       Assumes all values are positive when `template.is_amplitude` is `True`.

    .. rubric:: Slots

    **src**
        Input type float32 or complex64.
        Shape is number of rows by number of columns, where 5 percentiles
        are computed along the columns, per row.

    **dest**
        Output type float32.
        Shape is (5, number of rows of input)

    Parameters
    ----------
    template
        Operation template
    command_queue
        Command queue for the operation
    shape
        Shape of the source data
    column_range:
        Half-open interval of columns that will be processed. If not specified, all columns are
        processed.
    allocator
        Allocator used to allocate unbound slots
    """

    def __init__(
        self,
        template: Percentile5Template,
        command_queue: AbstractCommandQueue,
        shape: Tuple[int, int],
        column_range: Optional[Tuple[int, int]],
        allocator: Optional[accel.AbstractAllocator] = None,
    ) -> None:
        super().__init__(command_queue, allocator)
        if column_range is None:
            column_range = (0, shape[1])
        if column_range[1] <= column_range[0]:
            raise ValueError("column range is empty")
        if column_range[0] < 0 or column_range[1] > shape[1]:
            raise IndexError("column range is out of range")
        if column_range[1] - column_range[0] > template.max_columns:
            raise ValueError("columns exceeds max_columns")
        self.template = template
        self.kernel = template.program.get_kernel("percentile5_float")
        self.shape = shape
        self.column_range = column_range
        src_type = np.float32 if self.template.is_amplitude else np.complex64
        row_dim = accel.Dimension(shape[0], self.template.wgsy)
        col_dim = accel.Dimension(shape[1])
        self.slots["src"] = accel.IOSlot((row_dim, col_dim), src_type)
        self.slots["dest"] = accel.IOSlot((5, row_dim), np.float32)

    def _run(self) -> None:
        src = self.buffer("src")
        dest = self.buffer("dest")
        rows_padded = accel.roundup(src.shape[0], self.template.wgsy)
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                src.buffer,
                dest.buffer,
                np.int32(src.padded_shape[1]),
                np.int32(dest.padded_shape[1]),
                np.int32(self.column_range[0]),
                np.int32(self.column_range[1] - self.column_range[0]),
            ],
            global_size=(self.template.size, rows_padded),
            local_size=(self.template.size, self.template.wgsy),
        )

    def parameters(self) -> Mapping[str, Any]:
        return {
            "max_columns": self.template.max_columns,
            "is_amplitude": self.template.is_amplitude,
            "shape": self.slots["src"].shape,  # type: ignore
            "column_range": self.column_range,
        }



################################################################################
# Copyright (c) 2014-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for rank.mako."""

import os
from typing import Any

import numpy as np

try:
    from numpy.typing import ArrayLike, DTypeLike
except ImportError:
    DTypeLike = Any  # type: ignore
    ArrayLike = Any  # type: ignore
import pytest

from katsdpsigproc import accel
from katsdpsigproc.abc import AbstractCommandQueue, AbstractContext, AbstractProgram
from katsdpsigproc.accel import DeviceArray, build

_wgs = 128
_M = 1000
_N = 2000


@pytest.fixture(scope="module")
def data() -> np.ndarray:
    rs = np.random.RandomState(seed=1)  # Fixed seed to make test repeatable
    return rs.randint(0, _M, size=_N).astype(np.int32)


@pytest.fixture(scope="module")
def expected(data: np.ndarray) -> np.ndarray:
    expected = np.empty(_M, dtype=np.int32)
    for i in range(_M):
        expected[i] = np.sum(np.less(data, i))
    return expected


@pytest.fixture
def program(context: AbstractContext) -> AbstractProgram:
    return build(
        context,
        "test_rank.mako",
        {"size": _wgs},
        extra_dirs=[os.path.dirname(__file__)],
    )


@pytest.mark.parametrize(
    "kernel_name, wgs", [("test_rank_serial", 1), ("test_rank_parallel", _wgs)]
)
def test_rank(
    context: AbstractContext,
    command_queue: AbstractCommandQueue,
    data: np.ndarray,
    expected: np.ndarray,
    program: AbstractProgram,
    kernel_name: str,
    wgs: int,
) -> None:
    data_d = DeviceArray(context, shape=data.shape, dtype=data.dtype)
    data_d.set(command_queue, data)
    out_d = DeviceArray(context, shape=expected.shape, dtype=expected.dtype)
    kernel = program.get_kernel(kernel_name)
    command_queue.enqueue_kernel(
        kernel,
        [data_d.buffer, out_d.buffer, np.int32(_N), np.int32(_M)],
        global_size=(wgs,),
        local_size=(wgs,),
    )
    out = out_d.get(command_queue)
    np.testing.assert_equal(expected, out)


def run_float_func(
    context: AbstractContext,
    command_queue: AbstractCommandQueue,
    program: AbstractProgram,
    kernel_name: str,
    data: np.ndarray,
    output_size: int,
) -> np.ndarray:
    """Do common testing for find_min_float, find_max_float, median_non_zero_float."""
    size = 128  # Number of workitems
    kernel = program.get_kernel(kernel_name)
    data = data.astype(np.float32)
    N = data.shape[0]
    data_d = accel.DeviceArray(context, shape=data.shape, dtype=np.float32)
    data_d.set(command_queue, data)
    out_d = accel.DeviceArray(context, shape=(output_size,), dtype=np.float32)
    command_queue.enqueue_kernel(
        kernel,
        [data_d.buffer, out_d.buffer, np.int32(N)],
        global_size=(size,),
        local_size=(size,),
    )
    out = out_d.empty_like()
    out_d.get(command_queue, out)
    return out


class TestFindMinMaxFloat:
    def check_array(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue,
        program: AbstractProgram,
        data_: ArrayLike,
    ) -> None:
        data = np.asarray(data_, dtype=np.float32)
        expected = [np.nanmin(data), np.nanmax(data)]
        out = run_float_func(context, command_queue, program, "test_find_min_max_float", data, 2)
        assert expected[0] == out[0]
        assert expected[1] == out[1]

    @pytest.mark.parametrize("data", [[5.3], [-10.0, 5.5, np.nan, -20.0, np.nan]])
    def test_simple(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue,
        program: AbstractProgram,
        data: ArrayLike,
    ) -> None:
        self.check_array(context, command_queue, program, data)

    @pytest.mark.parametrize("ordered", [False, True])
    def test_random(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue,
        program: AbstractProgram,
        ordered: bool,
    ) -> None:
        rs = np.random.RandomState(seed=1)  # Fixed seed to make test repeatable
        data = rs.uniform(-10.0, 10.0, 1000)
        if ordered:
            data = np.sort(data)
        self.check_array(context, command_queue, program, data)

    def test_all_nan(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue,
        program: AbstractProgram,
    ) -> None:
        # Can't use check_array, because np.nanmin warns if all are NaN,
        # and asserting == doesn't handle NaN.
        data = np.array([np.nan, np.nan], dtype=np.float32)
        out = run_float_func(context, command_queue, program, "test_find_min_max_float", data, 2)
        assert np.isnan(out[0])
        assert np.isnan(out[1])


class TestMedianNonZero:
    def check_array(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue,
        program: AbstractProgram,
        data_: ArrayLike,
    ) -> None:
        data = np.asarray(data_, dtype=np.float32)
        expected = np.median(data[data > 0.0])
        out = run_float_func(context, command_queue, program, "test_median_non_zero", data, 2)
        assert expected == out[0]
        assert expected == out[1]

    @pytest.mark.parametrize("data", [[5.3], [1.2, 2.3, 3.4, 4.5], [0.0, 0.0, 0.0, 1.2, 5.3, 2.4]])
    def test_simple(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue,
        program: AbstractProgram,
        data: ArrayLike,
    ) -> None:
        self.check_array(context, command_queue, program, data)

    def test_big_even(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue,
        program: AbstractProgram,
    ) -> None:
        rs = np.random.RandomState(seed=1)  # Fixed seed to make test repeatable
        data = rs.random_sample(10001) + 0.5
        data[123] = 0.0
        self.check_array(context, command_queue, program, data)

    def test_big_odd(
        self,
        context: AbstractContext,
        command_queue: AbstractCommandQueue,
        program: AbstractProgram,
    ) -> None:
        rs = np.random.RandomState(seed=2)  # Fixed seed to make test repeatable
        data = rs.random_sample(10000) + 0.5
        data[456] = 0.0
        self.check_array(context, command_queue, program, data)



################################################################################
# Copyright (c) 2014-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpsigproc.tune`."""

import sqlite3
import threading
from typing import Any, Generator, Mapping, NoReturn
from unittest import mock

import pytest

from katsdpsigproc import tune
from katsdpsigproc.abc import AbstractContext


def test_autotune_basic() -> None:
    received = []
    received_lock = threading.Lock()

    def generate(a, b):
        with received_lock:
            received.append((a, b))
        return lambda iters: a * b

    best = tune.autotune(generate, time_limit=0.001, a=[1, 2], b=[7, 3])
    # Autotuning is parallel, so we can't assert anything about the order
    assert sorted(received) == [(1, 3), (1, 7), (2, 3), (2, 7)]
    assert best == {"a": 1, "b": 3}


def test_autotune_empty() -> None:
    with pytest.raises(ValueError):
        tune.autotune(lambda x, y: lambda iters: 0, x=[1, 2], y=[])


class CustomError(RuntimeError):
    pass


def test_autotune_some_raise() -> None:
    def generate(x):
        if x == 1:
            raise CustomError("x = 1")

        def measure(iters):
            if x == 3:
                raise CustomError("x = 3")
            return -x

        return measure

    best = tune.autotune(generate, x=[0, 1, 2, 3])
    assert best == {"x": 2}


def generate_raise(x: Any) -> NoReturn:
    raise CustomError(f"x = {x}")


def test_autotune_all_raise() -> None:
    with pytest.raises(CustomError, match=r"^x = 3$") as exc_info:
        tune.autotune(generate_raise, x=[1, 2, 3])
    # Check that the traceback refers to the original site, not
    # where it was re-raised
    assert exc_info.traceback[-1].name == "generate_raise"


class TestAutotuner:
    """Tests for the `autotuner` decorator.

    We use mocking to substitute an in-memory database, which is made to
    persist for the test instead of being closed each time the decorator is
    called.
    """

    autotune_mock = mock.Mock()

    @pytest.fixture
    def conn(self) -> Generator[sqlite3.Connection, None, None]:
        conn = sqlite3.connect(":memory:")
        yield conn
        conn.close()

    @pytest.fixture(autouse=True)
    def _prepare_autotune_mock(self) -> Generator[None, None, None]:
        # The autotune method has to be a class method due to magic in the
        # autotuner decorator, and hence autotune_mock has to be at class
        # scope. But we want it reset for each test.
        self.autotune_mock.reset()
        self.autotune_mock.side_effect = None
        yield
        self.autotune_mock.reset()

    @classmethod
    @tune.autotuner(test={"a": 3, "b": -1})
    def autotune(cls, context: AbstractContext, param: str) -> Mapping[str, Any]:
        return cls.autotune_mock(context, param)

    @classmethod
    @tune.autotuner(test={"a": 3, "b": -1})
    def autotune_no_args(cls, context: AbstractContext) -> Any:
        return cls.autotune_mock(context)

    @mock.patch("katsdpsigproc.tune._close_db")
    @mock.patch("katsdpsigproc.tune._open_db")
    @mock.patch("katsdpsigproc.tune.KATSDPSIGPROC_TUNE_MATCH", "exact")
    def test_autotune_match(
        self,
        open_db_mock: mock.Mock,
        close_db_mock: mock.Mock,
        conn: sqlite3.Connection,
    ) -> None:
        """Test behaviour of both autotune matching options ('exact' and 'nearest')."""
        open_db_mock.return_value = conn
        context = mock.NonCallableMock()

        # Check that autotuning is only performed once for a particular
        # context+arg_parameters set
        tuning_1 = {"a": 1, "b": 2}
        self.autotune_mock.return_value = tuning_1
        context.device.driver_version = "mock version"
        context.device.platform_name = "mock platform"
        context.device.name = "mock device"
        ret = self.autotune(context, "xyz")
        ret_2 = self.autotune(context, "xyz")
        self.autotune_mock.assert_called_once_with(context, "xyz")
        assert ret == ret_2 == tuning_1

        # Check that using a different context+arg_parameters set causes tuning to be
        # re-run. This also populates the database with another set
        tuning_2 = {"a": 3, "b": 4}
        self.autotune_mock.return_value = tuning_2
        context.device.driver_version = "another version"
        context.device.platform_name = "another platform"
        context.device.name = "another device"
        ret = self.autotune(context, "xyz")
        assert ret == tuning_2

        # Now test incremental fallback on device  [ driver version | platform | name ]
        # when KATSDPSIGPROC_TUNE_MATCH is set to 'nearest'
        with mock.patch("katsdpsigproc.tune.KATSDPSIGPROC_TUNE_MATCH", "nearest"):
            # autotuning re-runs if arg_parameters are different
            tuning_4 = {"a": 7, "b": 8}
            self.autotune_mock.return_value = tuning_4
            context.device.driver_version = "mock version"
            context.device.platform_name = "mock platform"
            context.device.name = "mock device"
            ret = self.autotune(context, "zzz")
            assert ret == tuning_4

            # Re-run with different configurations making sure the cached values are returned if
            # they exist in any configuration, and that autotuning does not run.
            self.autotune_mock.side_effect = RuntimeError

            context.device.driver_version = "abc"
            context.device.platform_name = "mock platform"
            context.device.name = "mock device"
            ret = self.autotune(context, "xyz")
            assert ret == tuning_1

            context.device.driver_version = "abc"
            context.device.platform_name = "abc"
            context.device.name = "another device"
            ret = self.autotune(context, "xyz")
            assert ret == tuning_2

            context.device.driver_version = "mock version"
            context.device.platform_name = "mock platform"
            context.device.name = "mock device"
            ret = self.autotune(context, "xyz")
            assert ret == tuning_1

            # If there is no match in the database, any tuning (for that parameter set) will
            # suffice
            context.device.driver_version = "x"
            context.device.platform_name = "y"
            context.device.name = "z"
            ret = self.autotune(context, "xyz")
            assert ret == tuning_1 or ret == tuning_2

    @mock.patch("katsdpsigproc.tune._close_db")
    @mock.patch("katsdpsigproc.tune._open_db")
    def test_autotune_empty_arg(
        self,
        open_db_mock: mock.Mock,
        close_db_mock: mock.Mock,
        conn: sqlite3.Connection,
    ) -> None:
        """Test behaviour of empty parameter sets."""
        open_db_mock.return_value = conn
        context = mock.NonCallableMock()

        tuning = {"a": 1, "b": 2}
        self.autotune_mock.return_value = tuning
        context.device.driver_version = "mock version"
        context.device.platform_name = "mock platform"
        context.device.name = "mock device"

        ret = self.autotune_no_args(context)
        assert ret == tuning



"""Reduction algorithms."""

from typing import Any, Callable, Mapping, Optional, Tuple, cast

import numpy as np
from typing_extensions import TypedDict

try:
    from numpy.typing import DTypeLike
except ImportError:
    DTypeLike = Any  # type: ignore

from . import accel, tune
from .abc import AbstractCommandQueue, AbstractContext


class _TuningDict(TypedDict):
    wgsx: int
    wgsy: int


class HReduceTemplate:
    """Performs reduction along rows in a 2D array.

    Only commutative reduction operators are supported.

    Parameters
    ----------
    context
        Context for which kernels will be compiled
    dtype
        Type of data elements
    ctype
        Type (in C/CUDA, not numpy) of data elements
    op
        C expression to combine two variables, *a* and *b*
    identity
        C expression for an identity value for *op*
    extra_code
        Arbitrary C code to paste in (for use by *op* or *identity*)
    tuning
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - wgsx: number of workitems per workgroup per row
        - wgsy: number of rows to handle per workgroup
    """

    def __init__(
        self,
        context: AbstractContext,
        dtype: DTypeLike,
        ctype: str,
        op: str,
        identity: str,
        extra_code: str = "",
        tuning: Optional[_TuningDict] = None,
    ) -> None:
        self.context = context
        self.dtype = dtype
        self.ctype = ctype
        if tuning is None:
            tuning = self.autotune(context, dtype, ctype, op, identity, extra_code)
        self.wgsx = tuning["wgsx"]
        self.wgsy = tuning["wgsy"]
        self.op = op
        self.identity = identity
        self.extra_code = extra_code
        self.program = accel.build(
            context,
            "hreduce.mako",
            {
                "wgsx": self.wgsx,
                "wgsy": self.wgsy,
                "type": self.ctype,
                "op": op,
                "identity": identity,
                "extra_code": extra_code,
            },
        )

    @classmethod
    @tune.autotuner(test={"wgsx": 64, "wgsy": 4})
    def autotune(
        cls,
        context: AbstractContext,
        dtype: DTypeLike,
        ctype: str,
        op: str,
        identity: str,
        extra_code: str,
    ) -> _TuningDict:
        queue = context.create_tuning_command_queue()
        shape = (2048, 1024)
        src = accel.DeviceArray(context, shape, dtype=dtype)
        dest = accel.DeviceArray(context, (shape[0],), dtype=dtype)

        def generate(wgsx: int, wgsy: int) -> Callable[[int], float]:
            wgs = wgsx * wgsy
            if wgs < 32 or wgs > 1024:
                raise RuntimeError(f"Skipping work group size {wgsx}x{wgsy}")
            template = cls(
                context,
                dtype,
                ctype,
                op,
                identity,
                extra_code,
                {"wgsx": wgsx, "wgsy": wgsy},
            )
            fn = template.instantiate(queue, shape)
            fn.bind(src=src, dest=dest)
            fn.ensure_all_bound()
            return tune.make_measure(queue, fn)

        return cast(
            _TuningDict,
            tune.autotune(generate, wgsx=[32, 64, 128], wgsy=[1, 2, 4, 8, 16]),
        )

    def instantiate(
        self,
        command_queue: AbstractCommandQueue,
        shape: Tuple[int, int],
        column_range: Optional[Tuple[int, int]] = None,
        allocator: Optional[accel.AbstractAllocator] = None,
    ) -> "HReduce":
        return HReduce(self, command_queue, shape, column_range, allocator)


class HReduce(accel.Operation):
    """Concrete instance of :class:`HReduceTemplate`.

    In each row, the elements in the specified column range are reduced using
    the reduction operator supplied to the template.

    .. rubric:: Slots

    **src** : *rows* × *columns*
        Input data
    **dest** : *rows*
        Output reductions

    Parameters
    ----------
    template
        Operation template
    command_queue
        Command queue for the operation
    shape
        Shape for the source slot
    column_range
        Half-open range of columns to reduce (defaults to the entire array)
    allocator
        Allocator used to allocate unbound slots
    """

    def __init__(
        self,
        template: HReduceTemplate,
        command_queue: AbstractCommandQueue,
        shape: Tuple[int, int],
        column_range: Optional[Tuple[int, int]] = None,
        allocator: Optional[accel.AbstractAllocator] = None,
    ) -> None:
        if len(shape) != 2:
            raise ValueError("shape must be 2-dimensional")
        if column_range is None:
            column_range = (0, shape[1])
        if column_range[0] < 0 or column_range[1] > shape[1]:
            raise ValueError("column range overflows the array")
        if column_range[0] >= column_range[1]:
            raise ValueError("column range is empty")

        super().__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel("hreduce")
        self.column_range = column_range
        self.slots["src"] = accel.IOSlot(
            (accel.Dimension(shape[0], self.template.wgsy), shape[1]),
            self.template.dtype,
        )
        self.slots["dest"] = accel.IOSlot(
            (accel.Dimension(shape[0], self.template.wgsy),), self.template.dtype
        )

    def _run(self) -> None:
        src = self.buffer("src")
        dest = self.buffer("dest")
        rows_padded = accel.roundup(src.shape[0], self.template.wgsy)
        n_columns = self.column_range[1] - self.column_range[0]
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                src.buffer,
                dest.buffer,
                np.int32(self.column_range[0]),
                np.int32(n_columns),
                np.int32(src.padded_shape[1]),
            ],
            global_size=(self.template.wgsx, rows_padded),
            local_size=(self.template.wgsx, self.template.wgsy),
        )

    def parameters(self) -> Mapping[str, Any]:
        return {
            "dtype": self.template.dtype,
            "ctype": self.template.ctype,
            "shape": self.slots["src"].shape,  # type: ignore
            "column_range": self.column_range,
            "op": self.template.op,
            "identity": self.template.identity,
            "extra_code": self.template.extra_code,
        }
